@@ -27,6 +27,7 @@ module Yesod.Form.Fields
     , Textarea (..)
     , boolField
     , checkBoxField
+    , fileField
       -- * File 'AForm's
     , fileAFormReq
     , fileAFormOpt
@@ -47,38 +48,38 @@ module Yesod.Form.Fields
 
 import Yesod.Form.Types
 import Yesod.Form.I18n.English
+import Yesod.Form.Functions (parseHelper)
 import Yesod.Handler (getMessageRender)
 import Yesod.Widget (toWidget, whamlet, GWidget)
 import Yesod.Message (RenderMessage (renderMessage), SomeMessage (..))
 import Text.Hamlet
-#if MIN_VERSION_blaze_html(0, 5, 0)
 import Text.Blaze (ToMarkup (toMarkup), preEscapedToMarkup, unsafeByteString)
 #define ToHtml ToMarkup
 #define toHtml toMarkup
 #define preEscapedText preEscapedToMarkup
-#else
-import Text.Blaze (ToHtml (..), preEscapedText, unsafeByteString)
-#endif
 import Text.Cassius
 import Data.Time (Day, TimeOfDay(..))
 import qualified Text.Email.Validate as Email
+#if MIN_VERSION_email_validate(1, 0, 0)
+import Data.Text.Encoding (encodeUtf8, decodeUtf8With)
+import Data.Text.Encoding.Error (lenientDecode)
+#endif
 import Network.URI (parseURI)
 import Database.Persist (PersistField)
 import Database.Persist.Store (Entity (..))
 import Text.HTML.SanitizeXSS (sanitizeBalance)
 import Control.Monad (when, unless)
-import Data.Maybe (listToMaybe)
+import Data.Maybe (listToMaybe, fromMaybe)
 
 import qualified Blaze.ByteString.Builder.Html.Utf8 as B
 import Blaze.ByteString.Builder (writeByteString, toLazyByteString)
 import Blaze.ByteString.Builder.Internal.Write (fromWriteList)
 import Database.Persist.Store (PersistEntityBackend)
-
-#if MIN_VERSION_blaze_html(0, 5, 0)
-import Text.Blaze.Html.Renderer.String (renderHtml)
-#else
-import Text.Blaze.Renderer.String (renderHtml)
+#if MIN_VERSION_persistent(1, 1, 0)
+import Database.Persist.Store (PersistMonadBackend)
 #endif
+
+import Text.Blaze.Html.Renderer.String (renderHtml)
 import qualified Data.ByteString as S
 import qualified Data.ByteString.Lazy as L
 import Data.Text (Text, unpack, pack)
@@ -92,27 +93,26 @@ import Yesod.Core (toPathPiece, GHandler, PathPiece, fromPathPiece)
 import Yesod.Persist (selectList, runDB, Filter, SelectOpt, YesodPersistBackend, Key, YesodPersist, PersistEntity, PersistQuery)
 import Control.Arrow ((&&&))
 
-import Control.Applicative ((<$>))
+import Control.Applicative ((<$>), (<|>))
+
+import Data.Attoparsec.Text (Parser, char, string, digit, skipSpace, endOfInput, parseOnly)
 
 defaultFormMessage :: FormMessage -> Text
 defaultFormMessage = englishFormMessage
 
-blank :: (Monad m, RenderMessage master FormMessage)
-      => (Text -> Either FormMessage a) -> [Text] -> m (Either (SomeMessage master) (Maybe a))
-blank _ [] = return $ Right Nothing
-blank _ ("":_) = return $ Right Nothing
-blank f (x:_) = return $ either (Left . SomeMessage) (Right . Just) $ f x
 
 intField :: (Integral i, RenderMessage master FormMessage) => Field sub master i
 intField = Field
-    { fieldParse = blank $ \s ->
+    { fieldParse = parseHelper $ \s ->
         case Data.Text.Read.signed Data.Text.Read.decimal s of
             Right (a, "") -> Right a
             _ -> Left $ MsgInvalidInteger s
 
     , fieldView = \theId name attrs val isReq -> toWidget [hamlet|
+$newline never
 <input id="#{theId}" name="#{name}" *{attrs} type="number" :isReq:required="" value="#{showVal val}">
 |]
+    , fieldEnctype = UrlEncoded
     }
   where
     showVal = either id (pack . showI)
@@ -120,32 +120,38 @@ intField = Field
 
 doubleField :: RenderMessage master FormMessage => Field sub master Double
 doubleField = Field
-    { fieldParse = blank $ \s ->
+    { fieldParse = parseHelper $ \s ->
         case Data.Text.Read.double s of
             Right (a, "") -> Right a
             _ -> Left $ MsgInvalidNumber s
 
     , fieldView = \theId name attrs val isReq -> toWidget [hamlet|
+$newline never
 <input id="#{theId}" name="#{name}" *{attrs} type="text" :isReq:required="" value="#{showVal val}">
 |]
+    , fieldEnctype = UrlEncoded
     }
   where showVal = either id (pack . show)
 
 dayField :: RenderMessage master FormMessage => Field sub master Day
 dayField = Field
-    { fieldParse = blank $ parseDate . unpack
+    { fieldParse = parseHelper $ parseDate . unpack
     , fieldView = \theId name attrs val isReq -> toWidget [hamlet|
+$newline never
 <input id="#{theId}" name="#{name}" *{attrs} type="date" :isReq:required="" value="#{showVal val}">
 |]
+    , fieldEnctype = UrlEncoded
     }
   where showVal = either id (pack . show)
 
 timeField :: RenderMessage master FormMessage => Field sub master TimeOfDay
 timeField = Field
-    { fieldParse = blank $ parseTime . unpack
+    { fieldParse = parseHelper parseTime
     , fieldView = \theId name attrs val isReq -> toWidget [hamlet|
+$newline never
 <input id="#{theId}" name="#{name}" *{attrs} :isReq:required="" value="#{showVal val}">
 |]
+    , fieldEnctype = UrlEncoded
     }
   where
     showVal = either id (pack . show . roundFullSeconds)
@@ -156,18 +162,19 @@ timeField = Field
 
 htmlField :: RenderMessage master FormMessage => Field sub master Html
 htmlField = Field
-    { fieldParse = blank $ Right . preEscapedText . sanitizeBalance
+    { fieldParse = parseHelper $ Right . preEscapedText . sanitizeBalance
     , fieldView = \theId name attrs val _isReq -> toWidget [hamlet|
-$# FIXME: There was a class="html" attribute, for what purpose?
+$newline never
 <textarea id="#{theId}" name="#{name}" *{attrs}>#{showVal val}
 |]
+    , fieldEnctype = UrlEncoded
     }
   where showVal = either id (pack . renderHtml)
 
 -- | A newtype wrapper around a 'Text' that converts newlines to HTML
 -- br-tags.
 newtype Textarea = Textarea { unTextarea :: Text }
-    deriving (Show, Read, Eq, PersistField)
+    deriving (Show, Read, Eq, PersistField, Ord)
 instance ToHtml Textarea where
     toHtml =
         unsafeByteString
@@ -184,36 +191,44 @@ instance ToHtml Textarea where
 
 textareaField :: RenderMessage master FormMessage => Field sub master Textarea
 textareaField = Field
-    { fieldParse =  blank $ Right . Textarea
+    { fieldParse = parseHelper $ Right . Textarea
     , fieldView = \theId name attrs val _isReq -> toWidget [hamlet|
+$newline never
 <textarea id="#{theId}" name="#{name}" *{attrs}>#{either id unTextarea val}
 |]
+    , fieldEnctype = UrlEncoded
     }
 
 hiddenField :: (PathPiece p, RenderMessage master FormMessage)
             => Field sub master p
 hiddenField = Field
-    { fieldParse = blank $ maybe (Left MsgValueRequired) Right . fromPathPiece
+    { fieldParse = parseHelper $ maybe (Left MsgValueRequired) Right . fromPathPiece
     , fieldView = \theId name attrs val _isReq -> toWidget [hamlet|
+$newline never
 <input type="hidden" id="#{theId}" name="#{name}" *{attrs} value="#{either id toPathPiece val}">
 |]
+    , fieldEnctype = UrlEncoded
     }
 
 textField :: RenderMessage master FormMessage => Field sub master Text
 textField = Field
-    { fieldParse = blank $ Right
+    { fieldParse = parseHelper $ Right
     , fieldView = \theId name attrs val isReq ->
         [whamlet|
+$newline never
 <input id="#{theId}" name="#{name}" *{attrs} type="text" :isReq:required value="#{either id id val}">
 |]
+    , fieldEnctype = UrlEncoded
     }
 
 passwordField :: RenderMessage master FormMessage => Field sub master Text
 passwordField = Field
-    { fieldParse = blank $ Right
+    { fieldParse = parseHelper $ Right
     , fieldView = \theId name attrs val isReq -> toWidget [hamlet|
+$newline never
 <input id="#{theId}" name="#{name}" *{attrs} type="password" :isReq:required="" value="#{either id id val}">
 |]
+    , fieldEnctype = UrlEncoded
     }
 
 readMay :: Read a => String -> Maybe a
@@ -230,68 +245,107 @@ parseDate = maybe (Left MsgInvalidDay) Right
 replace :: Eq a => a -> a -> [a] -> [a]
 replace x y = map (\z -> if z == x then y else z)
 
-parseTime :: String -> Either FormMessage TimeOfDay
-parseTime (h2:':':m1:m2:[]) = parseTimeHelper ('0', h2, m1, m2, '0', '0')
-parseTime (h1:h2:':':m1:m2:[]) = parseTimeHelper (h1, h2, m1, m2, '0', '0')
-parseTime (h1:h2:':':m1:m2:' ':'A':'M':[]) =
-    parseTimeHelper (h1, h2, m1, m2, '0', '0')
-parseTime (h1:h2:':':m1:m2:' ':'P':'M':[]) =
-    let [h1', h2'] = show $ (read [h1, h2] :: Int) + 12
-    in parseTimeHelper (h1', h2', m1, m2, '0', '0')
-parseTime (h1:h2:':':m1:m2:':':s1:s2:[]) =
-    parseTimeHelper (h1, h2, m1, m2, s1, s2)
-parseTime _ = Left MsgInvalidTimeFormat
+parseTime :: Text -> Either FormMessage TimeOfDay
+parseTime = either (Left . fromMaybe MsgInvalidTimeFormat . readMay . drop 2 . dropWhile (/= ':')) Right . parseOnly timeParser
 
-parseTimeHelper :: (Char, Char, Char, Char, Char, Char)
-                -> Either FormMessage TimeOfDay
-parseTimeHelper (h1, h2, m1, m2, s1, s2)
-    | h < 0 || h > 23 = Left $ MsgInvalidHour $ pack [h1, h2]
-    | m < 0 || m > 59 = Left $ MsgInvalidMinute $ pack [m1, m2]
-    | s < 0 || s > 59 = Left $ MsgInvalidSecond $ pack [s1, s2]
-    | otherwise = Right $ TimeOfDay h m s
+timeParser :: Parser TimeOfDay
+timeParser = do
+    skipSpace
+    h <- hour
+    _ <- char ':'
+    m <- minsec MsgInvalidMinute
+    hasSec <- (char ':' >> return True) <|> return False
+    s <- if hasSec then minsec MsgInvalidSecond else return 0
+    skipSpace
+    isPM <-
+        (string "am" >> return (Just False)) <|>
+        (string "AM" >> return (Just False)) <|>
+        (string "pm" >> return (Just True)) <|>
+        (string "PM" >> return (Just True)) <|>
+        return Nothing
+    h' <-
+        case isPM of
+            Nothing -> return h
+            Just x
+                | h <= 0 || h > 12 -> fail $ show $ MsgInvalidHour $ pack $ show h
+                | h == 12 -> return $ if x then 12 else 0
+                | otherwise -> return $ h + (if x then 12 else 0)
+    skipSpace
+    endOfInput
+    return $ TimeOfDay h' m s
   where
-    h = read [h1, h2] -- FIXME isn't this a really bad idea?
-    m = read [m1, m2]
-    s = fromInteger $ read [s1, s2]
+    hour = do
+        x <- digit
+        y <- (return <$> digit) <|> return []
+        let xy = x : y
+        let i = read xy
+        if i < 0 || i >= 24
+            then fail $ show $ MsgInvalidHour $ pack xy
+            else return i
+    minsec :: Num a => (Text -> FormMessage) -> Parser a
+    minsec msg = do
+        x <- digit
+        y <- digit <|> fail (show $ msg $ pack [x])
+        let xy = [x, y]
+        let i = read xy
+        if i < 0 || i >= 60
+            then fail $ show $ msg $ pack xy
+            else return $ fromIntegral (i :: Int)
 
 emailField :: RenderMessage master FormMessage => Field sub master Text
 emailField = Field
-    { fieldParse = blank $
+    { fieldParse = parseHelper $
+#if MIN_VERSION_email_validate(1, 0, 0)
+        \s ->
+            case Email.canonicalizeEmail $ encodeUtf8 s of
+                Just e -> Right $ decodeUtf8With lenientDecode e
+                Nothing -> Left $ MsgInvalidEmail s
+#else
         \s -> if Email.isValid (unpack s)
                 then Right s
                 else Left $ MsgInvalidEmail s
+#endif
     , fieldView = \theId name attrs val isReq -> toWidget [hamlet|
+$newline never
 <input id="#{theId}" name="#{name}" *{attrs} type="email" :isReq:required="" value="#{either id id val}">
 |]
+    , fieldEnctype = UrlEncoded
     }
 
 type AutoFocus = Bool
 searchField :: RenderMessage master FormMessage => AutoFocus -> Field sub master Text
 searchField autoFocus = Field
-    { fieldParse = blank Right
+    { fieldParse = parseHelper Right
     , fieldView = \theId name attrs val isReq -> do
         [whamlet|\
+$newline never
 <input id="#{theId}" name="#{name}" *{attrs} type="search" :isReq:required="" :autoFocus:autofocus="" value="#{either id id val}">
 |]
         when autoFocus $ do
           -- we want this javascript to be placed immediately after the field
-          [whamlet|<script>if (!('autofocus' in document.createElement('input'))) {document.getElementById('#{theId}').focus();}|]
+          [whamlet|
+$newline never
+<script>if (!('autofocus' in document.createElement('input'))) {document.getElementById('#{theId}').focus();}
+|]
           toWidget [cassius|
-            #{theId}
+            ##{theId}
               -webkit-appearance: textfield
             |]
+    , fieldEnctype = UrlEncoded
     }
 
 urlField :: RenderMessage master FormMessage => Field sub master Text
 urlField = Field
-    { fieldParse = blank $ \s ->
+    { fieldParse = parseHelper $ \s ->
         case parseURI $ unpack s of
             Nothing -> Left $ MsgInvalidUrl s
             Just _ -> Right s
     , fieldView = \theId name attrs val isReq ->
         [whamlet|
+$newline never
 <input ##{theId} name=#{name} *{attrs} type=url :isReq:required value=#{either id id val}>
 |]
+    , fieldEnctype = UrlEncoded
     }
 
 selectFieldList :: (Eq a, RenderMessage master FormMessage, RenderMessage master msg) => [(msg, a)] -> Field sub master a
@@ -299,9 +353,18 @@ selectFieldList = selectField . optionsPairs
 
 selectField :: (Eq a, RenderMessage master FormMessage) => GHandler sub master (OptionList a) -> Field sub master a
 selectField = selectFieldHelper
-    (\theId name inside -> [whamlet|<select ##{theId} name=#{name}>^{inside}|]) -- outside
-    (\_theId _name isSel -> [whamlet|<option value=none :isSel:selected>_{MsgSelectNone}|]) -- onOpt
-    (\_theId _name attrs value isSel text -> [whamlet|<option value=#{value} :isSel:selected *{attrs}>#{text}|]) -- inside
+    (\theId name attrs inside -> [whamlet|
+$newline never
+<select ##{theId} name=#{name} *{attrs}>^{inside}
+|]) -- outside
+    (\_theId _name isSel -> [whamlet|
+$newline never
+<option value=none :isSel:selected>_{MsgSelectNone}
+|]) -- onOpt
+    (\_theId _name _attrs value isSel text -> [whamlet|
+$newline never
+<option value=#{value} :isSel:selected>#{text}
+|]) -- inside
 
 multiSelectFieldList :: (Eq a, RenderMessage master FormMessage, RenderMessage master msg) => [(msg, a)] -> Field sub master [a]
 multiSelectFieldList = multiSelectField . optionsPairs
@@ -310,10 +373,10 @@ multiSelectField :: (Eq a, RenderMessage master FormMessage)
                  => GHandler sub master (OptionList a)
                  -> Field sub master [a]
 multiSelectField ioptlist =
-    Field parse view
+    Field parse view UrlEncoded
   where
-    parse [] = return $ Right Nothing
-    parse optlist = do
+    parse [] _ = return $ Right Nothing
+    parse optlist _ = do
         mapopt <- olReadExternal <$> ioptlist
         case mapM mapopt optlist of
              Nothing -> return $ Left "Error parsing values"
@@ -323,6 +386,7 @@ multiSelectField ioptlist =
         opts <- fmap olOptions $ lift ioptlist
         let selOpts = map (id &&& (optselected val)) opts
         [whamlet|
+$newline never
             <select ##{theId} name=#{name} :isReq:required multiple *{attrs}>
                 $forall (opt, optsel) <- selOpts
                     <option value=#{optionExternalValue opt} :optsel:selected>#{optionDisplay opt}
@@ -336,22 +400,30 @@ radioFieldList = radioField . optionsPairs
 
 radioField :: (Eq a, RenderMessage master FormMessage) => GHandler sub master (OptionList a) -> Field sub master a
 radioField = selectFieldHelper
-    (\theId _name inside -> [whamlet|<div ##{theId}>^{inside}|])
+    (\theId _name _attrs inside -> [whamlet|
+$newline never
+<div ##{theId}>^{inside}
+|])
     (\theId name isSel -> [whamlet|
-<div>
-    <input id=#{theId}-none type=radio name=#{name} value=none :isSel:checked>
-    <label for=#{theId}-none>_{MsgSelectNone}
+$newline never
+<label .radio for=#{theId}-none>
+    <div>
+        <input id=#{theId}-none type=radio name=#{name} value=none :isSel:checked>
+        _{MsgSelectNone}
 |])
     (\theId name attrs value isSel text -> [whamlet|
-<div>
-    <input id=#{theId}-#{value} type=radio name=#{name} value=#{value} :isSel:checked *{attrs}>
-    <label for=#{theId}-#{value}>#{text}
+$newline never
+<label .radio for=#{theId}-#{value}>
+    <div>
+        <input id=#{theId}-#{value} type=radio name=#{name} value=#{value} :isSel:checked *{attrs}>
+        \#{text}
 |])
 
 boolField :: RenderMessage master FormMessage => Field sub master Bool
 boolField = Field
-      { fieldParse = return . boolParser
+      { fieldParse = \e _ -> return $ boolParser e
       , fieldView = \theId name attrs val isReq -> [whamlet|
+$newline never
   $if not isReq
       <input id=#{theId}-none *{attrs} type=radio name=#{name} value=none checked>
       <label for=#{theId}-none>_{MsgSelectNone}
@@ -363,6 +435,7 @@ boolField = Field
 <input id=#{theId}-no *{attrs} type=radio name=#{name} value=no :showVal not val:checked>
 <label for=#{theId}-no>_{MsgBoolNo}
 |]
+    , fieldEnctype = UrlEncoded
     }
   where
     boolParser [] = Right Nothing
@@ -370,6 +443,7 @@ boolField = Field
       "" -> Right Nothing
       "none" -> Right Nothing
       "yes" -> Right $ Just True
+      "on" -> Right $ Just True
       "no" -> Right $ Just False
       t -> Left $ SomeMessage $ MsgInvalidBool t
     showVal = either (\_ -> False)
@@ -383,16 +457,19 @@ boolField = Field
 --
 checkBoxField :: RenderMessage m FormMessage => Field s m Bool
 checkBoxField = Field
-    { fieldParse = return . checkBoxParser
+    { fieldParse = \e _ -> return $ checkBoxParser e
     , fieldView  = \theId name attrs val _ -> [whamlet|
+$newline never
 <input id=#{theId} *{attrs} type=checkbox name=#{name} value=yes :showVal id val:checked>
 |]
+    , fieldEnctype = UrlEncoded
     }
 
     where
         checkBoxParser [] = Right $ Just False
         checkBoxParser (x:_) = case x of
             "yes" -> Right $ Just True
+            "on" -> Right $ Just True
             _     -> Right $ Just False
 
         showVal = either (\_ -> False)
@@ -428,10 +505,16 @@ optionsEnum :: (Show a, Enum a, Bounded a) => GHandler sub master (OptionList a)
 optionsEnum = optionsPairs $ map (\x -> (pack $ show x, x)) [minBound..maxBound]
 
 optionsPersist :: ( YesodPersist master, PersistEntity a
+#if MIN_VERSION_persistent(1, 1, 0)
+                  , PersistQuery (YesodPersistBackend master (GHandler sub master))
+                  , PathPiece (Key a)
+                  , PersistEntityBackend a ~ PersistMonadBackend (YesodPersistBackend master (GHandler sub master))
+#else
                   , PersistQuery (YesodPersistBackend master) (GHandler sub master)
                   , PathPiece (Key (YesodPersistBackend master) a)
-                  , RenderMessage master msg
                   , PersistEntityBackend a ~ YesodPersistBackend master
+#endif
+                  , RenderMessage master msg
                   )
                => [Filter a] -> [SelectOpt a] -> (a -> msg) -> GHandler sub master (OptionList (Entity a))
 optionsPersist filts ords toDisplay = fmap mkOptionList $ do
@@ -445,17 +528,17 @@ optionsPersist filts ords toDisplay = fmap mkOptionList $ do
 
 selectFieldHelper
         :: (Eq a, RenderMessage master FormMessage)
-        => (Text -> Text -> GWidget sub master () -> GWidget sub master ())
+        => (Text -> Text -> [(Text, Text)] -> GWidget sub master () -> GWidget sub master ())
         -> (Text -> Text -> Bool -> GWidget sub master ())
         -> (Text -> Text -> [(Text, Text)] -> Text -> Bool -> Text -> GWidget sub master ())
         -> GHandler sub master (OptionList a) -> Field sub master a
 selectFieldHelper outside onOpt inside opts' = Field
-    { fieldParse = \x -> do
+    { fieldParse = \x _ -> do
         opts <- opts'
         return $ selectParser opts x
     , fieldView = \theId name attrs val isReq -> do
         opts <- fmap olOptions $ lift opts'
-        outside theId name $ do
+        outside theId name attrs $ do
             unless isReq $ onOpt theId name $ not $ render opts val `elem` map optionExternalValue opts
             flip mapM_ opts $ \opt -> inside
                 theId
@@ -464,6 +547,7 @@ selectFieldHelper outside onOpt inside opts' = Field
                 (optionExternalValue opt)
                 ((render opts val) == optionExternalValue opt)
                 (optionDisplay opt)
+    , fieldEnctype = UrlEncoded
     }
   where
     render _ (Left _) = ""
@@ -476,6 +560,18 @@ selectFieldHelper outside onOpt inside opts' = Field
                     Nothing -> Left $ SomeMessage $ MsgInvalidEntry x
                     Just y -> Right $ Just y
 
+fileField :: RenderMessage master FormMessage => Field sub master FileInfo
+fileField = Field
+    { fieldParse = \_ files -> return $
+        case files of
+            [] -> Right Nothing
+            file:_ -> Right $ Just file
+    , fieldView = \id' name attrs _ isReq -> toWidget [hamlet|
+            <input id=#{id'} name=#{name} *{attrs} type=file :isReq:required>
+        |]
+    , fieldEnctype = Multipart
+    }
+
 fileAFormReq :: RenderMessage master FormMessage => FieldSettings master -> AForm sub master FileInfo
 fileAFormReq fs = AForm $ \(master, langs) menvs ints -> do
     let (name, ints') =
@@ -490,15 +586,16 @@ fileAFormReq fs = AForm $ \(master, langs) menvs ints -> do
                 Nothing -> (FormMissing, Nothing)
                 Just (_, fenv) ->
                     case Map.lookup name fenv of
-                        Nothing ->
+                        Just (fi:_) -> (FormSuccess fi, Nothing)
+                        _ ->
                             let t = renderMessage master langs MsgValueRequired
                              in (FormFailure [t], Just $ toHtml t)
-                        Just fi -> (FormSuccess fi, Nothing)
     let fv = FieldView
             { fvLabel = toHtml $ renderMessage master langs $ fsLabel fs
             , fvTooltip = fmap (toHtml . renderMessage master langs) $ fsTooltip fs
             , fvId = id'
             , fvInput = [whamlet|
+$newline never
 <input type=file name=#{name} ##{id'} *{fsAttrs fs}>
 |]
             , fvErrors = errs
@@ -520,13 +617,14 @@ fileAFormOpt fs = AForm $ \(master, langs) menvs ints -> do
                 Nothing -> (FormMissing, Nothing)
                 Just (_, fenv) ->
                     case Map.lookup name fenv of
-                        Nothing -> (FormSuccess Nothing, Nothing)
-                        Just fi -> (FormSuccess $ Just fi, Nothing)
+                        Just (fi:_) -> (FormSuccess $ Just fi, Nothing)
+                        _ -> (FormSuccess Nothing, Nothing)
     let fv = FieldView
             { fvLabel = toHtml $ renderMessage master langs $ fsLabel fs
             , fvTooltip = fmap (toHtml . renderMessage master langs) $ fsTooltip fs
             , fvId = id'
             , fvInput = [whamlet|
+$newline never
 <input type=file name=#{name} ##{id'} *{fsAttrs fs}>
 |]
             , fvErrors = errs
